@@ -10,11 +10,13 @@
 
 #include "soup-address.h"
 #include "soup-auth.h"
+#include "soup-connection.h"
 #include "soup-enum-types.h"
 #include "soup-marshal.h"
 #include "soup-message.h"
 #include "soup-message-private.h"
 #include "soup-misc.h"
+#include "soup-socket.h"
 #include "soup-uri.h"
 
 /**
@@ -64,11 +66,12 @@
  * trying to do.
  *
  * As described in the #SoupMessageBody documentation, the
- * @request_body and @response_body %data fields will not necessarily
- * be filled in at all times. When they are filled in, they will be
- * terminated with a '\0' byte (which is not included in the %length),
- * so you can use them as ordinary C strings (assuming that you know
- * that the body doesn't have any other '\0' bytes).
+ * @request_body and @response_body <literal>data</literal> fields
+ * will not necessarily be filled in at all times. When they are
+ * filled in, they will be terminated with a '\0' byte (which is not
+ * included in the <literal>length</literal>), so you can use them as
+ * ordinary C strings (assuming that you know that the body doesn't
+ * have any other '\0' bytes).
  *
  * For a client-side #SoupMessage, @request_body's %data is usually
  * filled in right before libsoup writes the request to the network,
@@ -352,7 +355,7 @@ soup_message_class_init (SoupMessageClass *message_class)
 	 * (If you need to requeue a message--eg, after handling
 	 * authentication or redirection--it is usually better to
 	 * requeue it from a #SoupMessage::got_body handler rather
-	 * than a #SoupMessage::got_header handler, so that the
+	 * than a #SoupMessage::got_headers handler, so that the
 	 * existing HTTP connection can be reused.)
 	 **/
 	signals[GOT_HEADERS] =
@@ -422,24 +425,25 @@ soup_message_class_init (SoupMessageClass *message_class)
 	 * @type: the content type that we got from sniffing
 	 * @params: (element-type utf8 utf8): a #GHashTable with the parameters
 	 *
-	 * This signal is emitted after %got-headers, and before the
-	 * first %got-chunk. If content sniffing is disabled, or no
-	 * content sniffing will be performed, due to the sniffer
-	 * deciding to trust the Content-Type sent by the server, this
-	 * signal is emitted immediately after %got_headers, and @type
-	 * is %NULL.
+	 * This signal is emitted after #SoupMessage::got-headers, and
+	 * before the first #SoupMessage::got-chunk. If content
+	 * sniffing is disabled, or no content sniffing will be
+	 * performed, due to the sniffer deciding to trust the
+	 * Content-Type sent by the server, this signal is emitted
+	 * immediately after #SoupMessage::got-headers, and @type is
+	 * %NULL.
 	 *
 	 * If the #SoupContentSniffer feature is enabled, and the
-	 * sniffer decided to perform sniffing, the first %got_chunk
-	 * emission may be delayed, so that the sniffer has enough
-	 * data to correctly sniff the content. It notified the
-	 * library user that the content has been sniffed, and allows
-	 * it to change the header contents in the message, if
-	 * desired.
+	 * sniffer decided to perform sniffing, the first
+	 * #SoupMessage::got-chunk emission may be delayed, so that the
+	 * sniffer has enough data to correctly sniff the content. It
+	 * notified the library user that the content has been
+	 * sniffed, and allows it to change the header contents in the
+	 * message, if desired.
 	 *
 	 * After this signal is emitted, the data that was spooled so
 	 * that sniffing could be done is delivered on the first
-	 * emission of %got_chunk.
+	 * emission of #SoupMessage::got-chunk.
 	 *
 	 * Since: 2.27.3
 	 **/
@@ -746,7 +750,9 @@ set_property (GObject *object, guint prop_id,
 		if (priv->tls_certificate)
 			g_object_unref (priv->tls_certificate);
 		priv->tls_certificate = g_value_dup_object (value);
-		if (priv->tls_certificate && !priv->tls_errors)
+		if (priv->tls_errors)
+			priv->msg_flags &= ~SOUP_MESSAGE_CERTIFICATE_TRUSTED;
+		else if (priv->tls_certificate)
 			priv->msg_flags |= SOUP_MESSAGE_CERTIFICATE_TRUSTED;
 		break;
 	case PROP_TLS_ERRORS:
@@ -1006,8 +1012,8 @@ soup_message_wrote_body (SoupMessage *msg)
  * soup_message_got_informational:
  * @msg: a #SoupMessage
  *
- * Emits the %got_informational signal, indicating that the IO layer
- * read a complete informational (1xx) response for @msg.
+ * Emits the #SoupMessage::got_informational signal, indicating that
+ * the IO layer read a complete informational (1xx) response for @msg.
  **/
 void
 soup_message_got_informational (SoupMessage *msg)
@@ -1019,8 +1025,8 @@ soup_message_got_informational (SoupMessage *msg)
  * soup_message_got_headers:
  * @msg: a #SoupMessage
  *
- * Emits the %got_headers signal, indicating that the IO layer
- * finished reading the (non-informational) headers for @msg.
+ * Emits the #SoupMessage::got_headers signal, indicating that the IO
+ * layer finished reading the (non-informational) headers for @msg.
  **/
 void
 soup_message_got_headers (SoupMessage *msg)
@@ -1033,8 +1039,8 @@ soup_message_got_headers (SoupMessage *msg)
  * @msg: a #SoupMessage
  * @chunk: the newly-read chunk
  *
- * Emits the %got_chunk signal, indicating that the IO layer finished
- * reading a chunk of @msg's body.
+ * Emits the #SoupMessage::got_chunk signal, indicating that the IO
+ * layer finished reading a chunk of @msg's body.
  **/
 void
 soup_message_got_chunk (SoupMessage *msg, SoupBuffer *chunk)
@@ -1061,8 +1067,8 @@ got_body (SoupMessage *req)
  * soup_message_got_body:
  * @msg: a #SoupMessage
  *
- * Emits the %got_body signal, indicating that the IO layer finished
- * reading the body for @msg.
+ * Emits the #SoupMessage::got_body signal, indicating that the IO
+ * layer finished reading the body for @msg.
  **/
 void
 soup_message_got_body (SoupMessage *msg)
@@ -1080,7 +1086,7 @@ soup_message_got_body (SoupMessage *msg)
  * finished sniffing the content type for @msg. If content sniffing
  * will not be performed, due to the sniffer deciding to trust the
  * Content-Type sent by the server, this signal is emitted immediately
- * after %got_headers, with %NULL as @content_type.
+ * after #SoupMessage::got_headers, with %NULL as @content_type.
  **/
 void
 soup_message_content_sniffed (SoupMessage *msg, const char *content_type, GHashTable *params)
@@ -1098,6 +1104,11 @@ soup_message_content_sniffed (SoupMessage *msg, const char *content_type, GHashT
 void
 soup_message_restarted (SoupMessage *msg)
 {
+	SoupMessagePrivate *priv = SOUP_MESSAGE_GET_PRIVATE (msg);
+
+	if (priv->msg_flags & SOUP_MESSAGE_CAN_REBUILD)
+		soup_message_body_truncate (msg->request_body);
+
 	g_signal_emit (msg, signals[RESTARTED], 0);
 }
 
@@ -1385,7 +1396,6 @@ soup_message_cleanup_response (SoupMessage *req)
 		priv->decoders = g_slist_delete_link (priv->decoders, priv->decoders);
 	}
 	priv->msg_flags &= ~SOUP_MESSAGE_CONTENT_DECODED;
-	priv->msg_flags &= ~SOUP_MESSAGE_CERTIFICATE_TRUSTED;
 
 	req->status_code = SOUP_STATUS_NONE;
 	if (req->reason_phrase) {
@@ -1394,12 +1404,6 @@ soup_message_cleanup_response (SoupMessage *req)
 	}
 	priv->http_version = priv->orig_http_version;
 
-	if (priv->tls_certificate) {
-		g_object_unref (priv->tls_certificate);
-		priv->tls_certificate = NULL;
-	}
-	priv->tls_errors = 0;
-
 	g_object_notify (G_OBJECT (req), SOUP_MESSAGE_STATUS_CODE);
 	g_object_notify (G_OBJECT (req), SOUP_MESSAGE_REASON_PHRASE);
 	g_object_notify (G_OBJECT (req), SOUP_MESSAGE_HTTP_VERSION);
@@ -1411,17 +1415,21 @@ soup_message_cleanup_response (SoupMessage *req)
 /**
  * SoupMessageFlags:
  * @SOUP_MESSAGE_NO_REDIRECT: The session should not follow redirect
- * (3xx) responses received by this message.
+ *   (3xx) responses received by this message.
+ * @SOUP_MESSAGE_CAN_REBUILD: The caller will rebuild the request
+ *   body if the message is restarted; see
+ *   soup_message_body_set_accumulate() for more details.
  * @SOUP_MESSAGE_OVERWRITE_CHUNKS: Deprecated: equivalent to calling
- * soup_message_body_set_accumulate() on the incoming message body
- * (ie, %response_body for a client-side request), passing %FALSE.
+ *   soup_message_body_set_accumulate() on the incoming message body
+ *   (ie, #SoupMessage:response_body for a client-side request),
+ *   passing %FALSE.
  * @SOUP_MESSAGE_CONTENT_DECODED: Set by #SoupContentDecoder to
- * indicate that it has removed the Content-Encoding on a message (and
- * so headers such as Content-Length may no longer accurately describe
- * the body).
- * @SOUP_MESSAGE_CERTIFICATE_TRUSTED: if %TRUE after an https response
- * has been received, indicates that the server's SSL certificate is
- * trusted according to the session's CA.
+ *   indicate that it has removed the Content-Encoding on a message (and
+ *   so headers such as Content-Length may no longer accurately describe
+ *   the body).
+ * @SOUP_MESSAGE_CERTIFICATE_TRUSTED: if set after an https response
+ *   has been received, indicates that the server's SSL certificate is
+ *   trusted according to the session's CA.
  *
  * Various flags that can be set on a #SoupMessage to alter its
  * behavior.
@@ -1722,25 +1730,27 @@ soup_message_set_status_full (SoupMessage *msg,
  * call @allocator, which should return a #SoupBuffer. (See
  * #SoupChunkAllocator for additional details.) Libsoup will then read
  * data from the network into that buffer, and update the buffer's
- * %length to indicate how much data it read.
+ * <literal>length</literal> to indicate how much data it read.
  *
  * Generally, a custom chunk allocator would be used in conjunction
  * with soup_message_body_set_accumulate() %FALSE and
  * #SoupMessage::got_chunk, as part of a strategy to avoid unnecessary
  * copying of data. However, you cannot assume that every call to the
- * allocator will be followed by a call to your %got_chunk handler; if
- * an I/O error occurs, then the buffer will be unreffed without ever
- * having been used. If your buffer-allocation strategy requires
- * special cleanup, use soup_buffer_new_with_owner() rather than doing
- * the cleanup from the %got_chunk handler.
+ * allocator will be followed by a call to your
+ * #SoupMessage::got_chunk handler; if an I/O error occurs, then the
+ * buffer will be unreffed without ever having been used. If your
+ * buffer-allocation strategy requires special cleanup, use
+ * soup_buffer_new_with_owner() rather than doing the cleanup from the
+ * #SoupMessage::got_chunk handler.
  *
  * The other thing to remember when using non-accumulating message
- * bodies is that the buffer passed to the %got_chunk handler will be
- * unreffed after the handler returns, just as it would be in the
- * non-custom-allocated case. If you want to hand the chunk data off
- * to some other part of your program to use later, you'll need to ref
- * the #SoupBuffer (or its owner, in the soup_buffer_new_with_owner()
- * case) to ensure that the data remains valid.
+ * bodies is that the buffer passed to the #SoupMessage::got_chunk
+ * handler will be unreffed after the handler returns, just as it
+ * would be in the non-custom-allocated case. If you want to hand the
+ * chunk data off to some other part of your program to use later,
+ * you'll need to ref the #SoupBuffer (or its owner, in the
+ * soup_buffer_new_with_owner() case) to ensure that the data remains
+ * valid.
  **/
 void
 soup_message_set_chunk_allocator (SoupMessage *msg,
@@ -1770,7 +1780,7 @@ soup_message_set_chunk_allocator (SoupMessage *msg,
  * This disables the actions of #SoupSessionFeature<!-- -->s with the
  * given @feature_type (or a subclass of that type) on @msg, so that
  * @msg is processed as though the feature(s) hadn't been added to the
- * session. Eg, passing #SOUP_TYPE_PROXY_RESOLVER for @feature_type
+ * session. Eg, passing #SOUP_TYPE_PROXY_URI_RESOLVER for @feature_type
  * will disable proxy handling and cause @msg to be sent directly to
  * the indicated origin server, regardless of system proxy
  * configuration.
@@ -1815,6 +1825,8 @@ soup_message_disables_feature (SoupMessage *msg, gpointer feature)
 /**
  * soup_message_get_first_party:
  * @msg: a #SoupMessage
+ *
+ * Gets @msg's first-party #SoupURI
  * 
  * Returns: (transfer none): the @msg's first party #SoupURI
  * 
@@ -1864,6 +1876,34 @@ soup_message_set_first_party (SoupMessage *msg,
 	g_object_notify (G_OBJECT (msg), SOUP_MESSAGE_FIRST_PARTY);
 }
 
+void
+soup_message_set_https_status (SoupMessage *msg, SoupConnection *conn)
+{
+	SoupSocket *sock;
+
+	sock = conn ? soup_connection_get_socket (conn) : NULL;
+	if (sock && soup_socket_is_ssl (sock)) {
+		GTlsCertificate *certificate;
+		GTlsCertificateFlags errors;
+
+		g_object_get (sock,
+			      SOUP_SOCKET_TLS_CERTIFICATE, &certificate,
+			      SOUP_SOCKET_TLS_ERRORS, &errors,
+			      NULL);
+		g_object_set (msg,
+			      SOUP_MESSAGE_TLS_CERTIFICATE, certificate,
+			      SOUP_MESSAGE_TLS_ERRORS, errors,
+			      NULL);
+		if (certificate)
+			g_object_unref (certificate);
+	} else {
+		g_object_set (msg,
+			      SOUP_MESSAGE_TLS_CERTIFICATE, NULL,
+			      SOUP_MESSAGE_TLS_ERRORS, 0,
+			      NULL);
+	}
+}
+
 /**
  * soup_message_get_https_status:
  * @msg: a #SoupMessage
@@ -1895,3 +1935,38 @@ soup_message_get_https_status (SoupMessage           *msg,
 		*errors = priv->tls_errors;
 	return priv->tls_certificate != NULL;
 }
+
+/**
+ * soup_message_set_redirect:
+ * @msg: a #SoupMessage
+ * @status_code: a 3xx status code
+ * @redirect_uri: the URI to redirect @msg to
+ *
+ * Sets @msg's status_code to @status_code and adds a Location header
+ * pointing to @redirect_uri. Use this from a #SoupServer when you
+ * want to redirect the client to another URI.
+ *
+ * @redirect_uri can be a relative URI, in which case it is
+ * interpreted relative to @msg's current URI. In particular, if
+ * @redirect_uri is just a path, it will replace the path
+ * <emphasis>and query</emphasis> of @msg's URI.
+ *
+ * Since: 2.38
+ */
+void
+soup_message_set_redirect (SoupMessage *msg, guint status_code,
+			   const char *redirect_uri)
+{
+	SoupURI *location;
+	char *location_str;
+
+	location = soup_uri_new_with_base (soup_message_get_uri (msg), redirect_uri);
+	g_return_if_fail (location != NULL);
+
+	soup_message_set_status (msg, status_code);
+	location_str = soup_uri_to_string (location, FALSE);
+	soup_message_headers_replace (msg->response_headers, "Location",
+				      location_str);
+	g_free (location_str);
+	soup_uri_free (location);
+}
