@@ -3,22 +3,25 @@
  * Copyright (C) 2001-2003, Ximian, Inc.
  */
 
-#include "test-utils.h"
+#include "config.h"
 
+#include <ctype.h>
 #include <dirent.h>
-#include <errno.h>
 #include <fcntl.h>
+#include <errno.h>
+#include <signal.h>
 #include <stdio.h>
+#include <stdlib.h>
+#include <string.h>
 #include <sys/stat.h>
-
-#ifdef G_OS_WIN32
-#include <getopt.h>
-#endif
+#include <unistd.h>
 
 #ifdef HAVE_MMAP
 #include <sys/mman.h>
 #endif
 
+#include <libsoup/soup.h>
+
 #ifdef HAVE_MMAP
 struct mapping {
 	void   *start;
@@ -124,7 +127,6 @@ do_get (SoupServer *server, SoupMessage *msg, const char *path)
 			g_free (index_path);
 			return;
 		}
-		g_free (index_path);
 
 		listing = get_directory_listing (path);
 		soup_message_set_response (msg, "text/html",
@@ -221,13 +223,13 @@ server_callback (SoupServer *server, SoupMessage *msg,
 	SoupMessageHeadersIter iter;
 	const char *name, *value;
 
-	g_print ("%s %s HTTP/1.%d\n", msg->method, path,
-		 soup_message_get_http_version (msg));
+	printf ("%s %s HTTP/1.%d\n", msg->method, path,
+		soup_message_get_http_version (msg));
 	soup_message_headers_iter_init (&iter, msg->request_headers);
 	while (soup_message_headers_iter_next (&iter, &name, &value))
-		g_print ("%s: %s\n", name, value);
+		printf ("%s: %s\n", name, value);
 	if (msg->request_body->length)
-		g_print ("%s\n", msg->request_body->data);
+		printf ("%s\n", msg->request_body->data);
 
 	file_path = g_strdup_printf (".%s", path);
 
@@ -239,7 +241,7 @@ server_callback (SoupServer *server, SoupMessage *msg,
 		soup_message_set_status (msg, SOUP_STATUS_NOT_IMPLEMENTED);
 
 	g_free (file_path);
-	g_print ("  -> %d %s\n\n", msg->status_code, msg->reason_phrase);
+	printf ("  -> %d %s\n\n", msg->status_code, msg->reason_phrase);
 }
 
 static void
@@ -277,8 +279,8 @@ main (int argc, char **argv)
 			ssl_port = atoi (optarg);
 			break;
 		default:
-			g_printerr ("Usage: %s [-p port] [-c ssl-cert-file -k ssl-key-file [-s ssl-port]]\n",
-				    argv[0]);
+			fprintf (stderr, "Usage: %s [-p port] [-c ssl-cert-file -k ssl-key-file [-s ssl-port]]\n",
+				 argv[0]);
 			exit (1);
 		}
 	}
@@ -287,13 +289,13 @@ main (int argc, char **argv)
 				  SOUP_SERVER_SERVER_HEADER, "simple-httpd ",
 				  NULL);
 	if (!server) {
-		g_printerr ("Unable to bind to server port %d\n", port);
+		fprintf (stderr, "Unable to bind to server port %d\n", port);
 		exit (1);
 	}
 	soup_server_add_handler (server, NULL,
 				 server_callback, NULL, NULL);
-	g_print ("\nStarting Server on port %d\n",
-		 soup_server_get_port (server));
+	printf ("\nStarting Server on port %d\n",
+		soup_server_get_port (server));
 	soup_server_run_async (server);
 
 	if (ssl_cert_file && ssl_key_file) {
@@ -304,17 +306,17 @@ main (int argc, char **argv)
 			NULL);
 
 		if (!ssl_server) {
-			g_printerr ("Unable to bind to SSL server port %d\n", ssl_port);
+			fprintf (stderr, "Unable to bind to SSL server port %d\n", ssl_port);
 			exit (1);
 		}
 		soup_server_add_handler (ssl_server, NULL,
 					 server_callback, NULL, NULL);
-		g_print ("Starting SSL Server on port %d\n", 
-			 soup_server_get_port (ssl_server));
+		printf ("Starting SSL Server on port %d\n", 
+			soup_server_get_port (ssl_server));
 		soup_server_run_async (ssl_server);
 	}
 
-	g_print ("\nWaiting for requests...\n");
+	printf ("\nWaiting for requests...\n");
 
 	loop = g_main_loop_new (NULL, TRUE);
 	g_main_loop_run (loop);
